
/*
  DaisyWheelPrinter
  by Lee Schafer
  10-13-2021

  This program changes the Smith Corona PWP 78 DS daisy wheel typewriter (PWP = Personal Word Processor)
  into a printer connected to the 25-pin serial port of an IBM XT. Smith Corona quit making all
  typewriters in 2005. Other hobbyists have converted typewriters by mounting a bank of solenoids to
  the keyboard. The solenoids depress the typewriter keys, but that prevents the user from typing
  on the keyboard again. The author's method uses an Arduino Nano Every to simulate the membrane keyboard,
  so that either a new RS232 port or the existing keyboard can input text.

  The Smith Corona keyboard does not output either serial or ASCII parallel codes. Instead, it forma an
  8 x 8 matrix which shorts 2 wires together when a key is pressed. (There are 61 keys on a PWP keyboard,
  allowing for 3 unused combinations in the 64 slot matrix. I employed one of the 3 (0x02) to indicate a
  non-printing character.)

  This keyboard membrane cable has pins 1 to 18. Pins 10 and 11 supply power to the Caps Lock LED. With
  no key pressed, pin 3 and pins 12 thru 18 are low (lets call these the 8 Low Lines), and pins 1, 2,
  and 4 thru 9 are high (we'll call these the 8 High Lines.) When any key is pressed, a High Line is
  shorted to low. Next, the typewriter puts a high pulse .5 mSec wide every 5 mSec on every Low Line,
  but pulses are offset so the typewriter can tell which Low Line is connected to the pressed key.
  From then on, the High Line follows whatever level the connected Low Line happens to be, until the
  key is released, when it returns to high.

  If Shift is held down while a letter is struck, two High Lines are involved: one for the Shift key and
  one for the letter. Depending on the letter, it is possible for both High Lines to be the same, so this
  one wire pulses high at two different times. Even stranger, this output line is shorted to two inputs 
  that could be in different states (one HIGH, one LOW), so what is the output of such a circuit? The 
  answer is LOW (ground takes precidence, found by experimentation). It's also possible that both High 
  Lines get their pulse from the same Low Line, which is not a problem.

  To make it clearer, this program calls the Low Lines L0 thru L7, and the High Lines H0 thru H7.
  Each value in the Flex array (each flexVal) is laid out as follows:
  Bit 7 = Shift bit (See below)
  Bits 6, 5, & 4 = select a high line H0 thru H7
  Bit 3 = Code bit (See below)
  Bits 2, 1, & 0 = select a low line L0 thru L7

  Special characxters [, ^, ~, `, Ç, and ] all require both Shift and Code-K. Bold, Center, SubScript,
  SuperScript, and Underline require only Code (not K, and not Shift). If flexVal has bit 3 (Code) set
  but bit 7 (Shift) is not set, the Code routine is executed, not Code-K. 
  
  ASCII characters missing fron the daisy wheel, including \, |, {, }, <, and > are all replaced by @. 
  The code for non-printing characters such as BELL is 0x02. Also, either Shift key will unlock CapsLock.
*/

// Global variables in decreasing order of size
unsigned long startTime;
int pinHi;                              // For H0 thru H7, letter or Shift or Code, whichever is first
int pinLo;                              // For L0 thru L7, letter or Shift or Code, whichever is first
int pinHi2;                             // For character High Lines (if second)
int pinLo2;                             // For character Low Lines (if second)
const int H0 = 2;                       // Count pins by their D number (D2). Don't count from pin 1.
const int H1 = 3;
const int H2 = 4;
const int H3 = 5;
const int H4 = 6;
const int H5 = 7;
const int H6 = 8;
const int H7 = 9;
const int PRINT = 10;
const int CTS = 11;
const int L0 = 14;                      // Pin 14 and higher are analog pins
const int L1 = 15;
const int L2 = 16;
const int L3 = 17;
const int L4 = 18;
const int L5 = 19;
const int L6 = 20;
const int L7 = 21;
byte flexVal; 
const byte flex [] = {0x02, 0x02, 0x50, 0x71, 0x70, 0x72, 0x02, 0x02, 0x15, 0x73, 0x02, 0x02, 0x78, 0x26, 0x29, 0x4D,
                      0x02, 0x02, 0x3D, 0x02, 0x08, 0xCD, 0xDD, 0xDC, 0xCC, 0xAC, 0x02, 0x02, 0xB6, 0x36, 0x94, 0x0F,
                      0x30, 0xE4, 0xC7, 0xC4, 0xB4, 0xA4, 0xF5, 0x47, 0xD5, 0xC5, 0xE5, 0xA5, 0x27, 0x35, 0x17, 0x60,
                      0x45, 0x64, 0x54, 0x44, 0x34, 0x24, 0x14, 0x75, 0x65, 0x55, 0xD7, 0x57, 0xD4, 0x25, 0xD4, 0xE0,
                      0xD4, 0xE2, 0xA1, 0xC1, 0xC2, 0xC3, 0xB2, 0xA2, 0x92, 0xE6, 0x96, 0xF7, 0xE7, 0xB7, 0x91, 0xD6,
                      0xC6, 0xE3, 0xB3, 0xD2, 0xA3, 0xF6, 0xB1, 0xD3, 0xD1, 0x93, 0xE1, 0xFD, 0xD4, 0xED, 0xEC, 0xB5,
                      0xBC, 0x62, 0x21, 0x41, 0x42, 0x43, 0x32, 0x22, 0x12, 0x66, 0x16, 0x77, 0x67, 0x37, 0x11, 0x56,
                      0x46, 0x63, 0x33, 0x52, 0x23, 0x76, 0x31, 0x53, 0x51, 0x13, 0x61, 0xD4, 0xD4, 0xD4, 0x9C, 0x10
                     };                    
byte ring[4096];
bool endKey = false;                     


void setup() {                          // the setup routine runs once when you press reset
  Serial1.begin(9600);
  pinMode(L0, INPUT);                   // A0
  pinMode(L1, INPUT);
  pinMode(L2, INPUT);
  pinMode(L3, INPUT);
  pinMode(L4, INPUT);
  pinMode(L5, INPUT);
  pinMode(L6, INPUT);
  pinMode(L7, INPUT);                   // A7
  pinMode(PRINT, INPUT);
  pinMode(CTS, OUTPUT);
  pinMode(LED_BUILTIN, OUTPUT);
  pinMode(H0, OUTPUT);                  // D2
  pinMode(H1, OUTPUT);
  pinMode(H2, OUTPUT);
  pinMode(H3, OUTPUT);
  pinMode(H4, OUTPUT);
  pinMode(H5, OUTPUT);
  pinMode(H6, OUTPUT);
  pinMode(H7, OUTPUT);                  // D9
}


void twoLines() {                       // Routine makes pinHi = whatever pinLo is
  digitalWrite(pinHi, LOW);             // Initially set pinHi LOW to indicate "key press"
  int linger = 0;
  unsigned long limit = 40;
  if (endKey == true) {
    limit = 80;                         // Loop for 40 mSec if continuing onto fourLines()
    linger = 30;
  }                                     // or 80 mSec if ending in twoLines()
  startTime = millis();
  do
  {
    if (digitalRead(pinLo) == HIGH) {
      digitalWrite(pinHi, HIGH);        // When pinLo goes HIGH, pinHi reads HIGH
    } else {
      digitalWrite(pinHi, LOW);         // This works for any number of times pinLo goes low
    }
  } while ((millis() - startTime) <= limit);    // 40 mSec to 500 mSec. Any longer than 500 mSec gets repeat-key
  if (endKey == 1) {
    digitalWrite(pinHi, HIGH);          // High line returns to HIGH if "key" is lifted
    delay(linger);                      // Between-letter delay = 0 Msec if fourlines() or 42 Msec
  }
}


void fourLines() {                      // This makes both High Lines follow their Low Lines
  digitalWrite(pinHi2, LOW);            // Initially set pinHi2 LOW to indicate "key press"
  startTime = millis();                 // pinHi is already set LOW in twolines()
  do
  {
    if (digitalRead(pinLo2) == HIGH) {
      digitalWrite(pinHi2, HIGH);
    } else {
      digitalWrite(pinHi2, LOW);        // This works for any number of times pinLo2 goes low
    }
    if (digitalRead(pinLo) == HIGH) {
      digitalWrite(pinHi, HIGH);        // When pinLo drops LOW, pinHi follows suit
    } else {
      digitalWrite(pinHi, LOW);
    }                                   // Total loop time for twoLines() and fourLines() is 80 mSec
  } while ((millis() - startTime) <= 40);    // 40 mSec to 500 mSec. Any longer gets repeat-key
  digitalWrite(pinHi2, HIGH);           // High lines always return to HIGH when "key" is lifted
  digitalWrite(pinHi, HIGH);            // Code key is the last to lift
  delay(30);                            // Between-character delay = 30 mSec
}


void getFlex() {                        // getFlex inputs "flexVal" & outputs pinHi & pinLo (all global)
  byte triVal = flexVal & 112;          // High nibble ANDed with bits 64, 32, & 16
  triVal = triVal >> 4;                 // Right-shift nibble by 4 bits
  pinHi = triVal + 2;                   // pinHi selects H0 thru H7 + 2 (D2 thru D9)
  triVal = flexVal & 7;                 // Low nibble ANDed with bits 4, 2, & 1
  pinLo = triVal + 14;                  // pinLo selects L0 thru L7 and adds 14 to choose analog pins
}


void normal() {
  // Set all High Lines to HIGH so keyboard can be used when RS232 is not signaling
  digitalWrite(H0, HIGH);
  digitalWrite(H1, HIGH);
  digitalWrite(H2, HIGH);
  digitalWrite(H3, HIGH);
  digitalWrite(H4, HIGH);
  digitalWrite(H5, HIGH);
  digitalWrite(H6, HIGH);
  digitalWrite(H7, HIGH);
}


void loop() {
  int head = 0;
  int tail = 0;
  byte code = 0;
  byte inChar;
  byte oldShift = 0;
  byte shift = 0;
  bool after = false;
  bool active = false;
  bool done = false;
  if (digitalRead(PRINT) == LOW) {       // Cable is not detected or Print switch is off
    digitalWrite(CTS, HIGH);             // CTS is de-asserted (MAX232 chip inverts this signal)
    digitalWrite(LED_BUILTIN, LOW);      // turn the LED off (LOW is the voltage level)
    normal();
    delay(500);                          // No serial data is coming during PRINT = LOW mode. Delay 1/2 second
  }
  else                                   // Cable is detected and Print switch is on
  {
    digitalWrite(LED_BUILTIN, HIGH);     // turn the LED on to show Print Mode
    digitalWrite(CTS, LOW);              // Assert CTS once (MAX232 inverts signal)
    while ((done == false) && (digitalRead(PRINT) == HIGH)) {   // LOW breaks out of the while loop
      if (Serial1.available() > 0) {     // } else ...
        startTime = millis(); 
        ring[tail] = Serial1.read();     // This puts data into array & also reduces count in Nano Every's buffer            
        tail += 1;                       // At 9600 baud, Nano Every's ring buffer fills up in 66.6 Msec
        if (tail > 4095) {
          tail = 0;
        }
        if (tail > head) {        
          if (tail - head > 4000) {      // If buffer count > 4000, de-assert CTS
            digitalWrite(CTS, HIGH);     // De-assert CTS
          }
        }
        if (head > tail) {               // If buffer count > 4000, de-assert CTS 
          if (((4095 - head) + tail) > 4000) {
            digitalWrite(CTS, HIGH);     // De-assert CTS
          }
        }
        active = true; 
        after = true;
      } else {
        active = false;
      }
      if (active == false && after == true) {  // Executes each time the while loop passes
        if ((millis() - startTime) >= 2000) {  // If no more data has arrived in 2 seconds, begin printing
          digitalWrite(CTS, HIGH);       // De-assert CTS (MAX232 inverts the signal)
          done = true;
          after = false;
        }
      }
    }                                    // while ...
    while (head != tail) {               // Begin typewriter printing until ring[4096] is empty
      inChar = ring[head];
      head += 1;
      if (head > 4095) {
        head = 0;
      }
      flexVal = flex[inChar];            // The rest of flexVal is decoded in getFlex
      code = bitRead(flexVal, 3);        // code is either 0 or 1
      shift = bitRead(flexVal, 7);       // shift is either 0 or 1
      
      if (code == 0 && shift == 0 && oldShift == 0) {                                     // LOWER CASE SECTION
        getFlex();                       // getFlex adds 14 to pinLo and 2 to pinHi
        endKey = true;
        twoLines();
        oldShift = 0;
        normal();
      }
      if (code == 0 && shift == 0 && oldShift == 1) {
        pinHi = 9;                       // Shift = 7 + 2 = 9     H7
        pinLo = 14;                      // Shift = 0 + 14 = 14   L0
        endKey = true;
        twoLines();                      // Either shift unlocks CapsLock
        oldShift = 0;
        getFlex();                       // getFlex adds 14 to pinLo and 2 to pinHi
        twoLines();
        normal();
      }
//    The author decided to comment out the part that handles foreign language symbols,
//    since financial software is not likely to use them. This reduces program size, 
//    but the full program is given here for those who want it.
//    Setting words to BOLD is still available in the CODE SECTION.       
//      if (code == 0 && shift == 0 && oldCode == 1) {                                          // CODE-K SECTION
//        if (oldShift == 1) {             // To use this section, oldCode variable must be initialized.
//          pinHi = 9;                     // Shift = 7 + 2 = 9     H7
//          pinLo = 14;                    // Shift = 0 + 14 = 14   L0
//          endKey = 1;
//          twoLines();                    // Either shift unlocks CapsLock
//          oldShift = 0;
//        }
//        getFlex();                       // getFlex adds 14 to pinLo and 2 to pinHi
//        endKey = 1;
//        twoLines();                      // Character must be printed before Keyboard II is disabled
//        normal();
//        pinHi = 7;                       // Code = 5 + 2 = 7    H5
//        pinLo = 14;                      // Code = 0 + 14 = 14  L0
//        pinHi2 = 9;                      // k = 7 + 2 = 9       H7
//        pinLo2 = 21;                     // k = 7 + 14 = 21     Lower-case 'k'
//        endKey = 0;
//        twoLines();
//        fourLines();                     // Disable Keyboard II. Doesn't use threeLines()
//        oldCode = 0;
//        normal();                
//      }
//      if (code == 1 && shift == 1 && oldCode == 1) {
//        if (oldShift == 0) {
//          pinHi = 9;                     // CapsLock = 7 + 2 = 9     H7
//          pinLo = 16;                    // CapsLock = 2 + 14 = 16   L2
//          endKey = 1;
//          twoLines();
//        }
//        oldCode = 1;
//        oldShift = 1;
//        getFlex();
//        endKey = 1;
//        twoLines();                      // Print French or Spanish character
//        normal(); 
//      }
//      if (code == 1 && shift == 1 && oldCode == 0) {
//        pinHi = 7;                       // Code = 5 + 2 = 7    H5
//        pinLo = 14;                      // Code = 0 + 14 = 14  L0
//        pinHi2 = 9;                      // k = 7 + 2 = 9       H7
//        pinLo2 = 21;                     // k = 7 + 14 = 21     Lower-case 'k'
//        endKey = 0;
//        twoLines();
//        fourLines();                     // Enable Keyboard II. Doesn't use threeLines()
//        normal();
//        if (oldShift == 0) {
//          pinHi = 9;                     // CapsLock = 7 + 2 = 9     H7
//          pinLo = 16;                    // CapsLock = 2 + 14 = 16   L2
//          endKey = 1;
//          twoLines();
//       }
//        oldCode = 1;
//        oldShift = 1;
//        getFlex();
//        endKey = 1;
//        twoLines();                      // None of the foreign language characters use threeLines()
//        normal();
//      }      
      if (code == 1 && shift == 0) {                                                            // CODE SECTION
        pinHi = 7;                       // Code = 5 + 2 = 7     H5
        pinLo = 14;                      // Code = 0 + 14 = 14   L0
        endKey = false;
        twoLines();
        getFlex();                       // getFlex adds 14 to pinLo and 2 to pinHi
        pinHi2 = pinHi;                  
        pinLo2 = pinLo;                  // getFlex over-writes pinHi & pinLo, so they must be re-installed
        pinHi = 7;                       // Code = 5 + 2 = 7     H5
        pinLo = 14;                      // Code = 0 + 14 = 14   L0
        fourLines();                     // No Bold, Underline, etc. command goes to threeLines()
        normal();
      }
      if (code == 0 && shift == 1 && oldShift == 1) {                                          // SHIFT SECTION
        getFlex();                       // getFlex adds 14 to pinLo and 2 to pinHi
        endKey = true;
        twoLines();
        oldShift = 1;
        normal();
      }
      if (code == 0 && shift == 1 && oldShift == 0) {
        pinHi = 9;                       // CapsLock = 7 + 2 = 9     H7
        pinLo = 16;                      // CapsLock = 2 + 14 = 16   L2
        endKey = true;
        twoLines();
        oldShift = 1;
        getFlex();                       // getFlex adds 14 to pinLo and 2 to pinHi
        twoLines();
        normal();
      }
    }                                    // while (head != tail) {
    if (head == tail) {
      done = false; 
    }    
  }                                      // else {
}                                        // void loop() {
